import asyncio
from typing import List

from cenao.app import AppFeature

from addr_agent.agent_feature.client import AddrClient
from addr_agent.agent_feature.hwreport import generate_hw_report
from addr_agent.agent_feature.mapping import RPCMethod, RPC_METHOD_MAPPING
from addr_agent.agent_feature.models import RPCTaskResult, RPCTaskStatus


class AgentFeature(AppFeature):
    NAME = 'agent'

    client: AddrClient

    fetch_tasks: bool = True
    poll_interval: int
    poll_task: asyncio.Task

    def on_init(self):
        host = self.config.get('host')
        if not host:
            raise Exception('Token must be configured')

        token = self.config.get('token')
        if not token:
            raise Exception('Token must be configured')

        self.client = AddrClient(self.logger, host, token)

        self.poll_interval = self.config.get('interval', 10)

    async def on_startup(self):
        await self.client.start()
        hw_report = generate_hw_report()
        await self.client.call_task(RPCMethod.ADDR_V1_HWREPORT, hw_report)
        self.poll_task = asyncio.create_task(self.poll())

    async def poll(self):
        self.logger.info("Start polling RPC tasks")
        try:
            while self.fetch_tasks:
                try:
                    task = await self.client.poll_task()
                    if task:
                        self.logger.info('Got RPC task %s', task)
                        try:
                            method = RPCMethod(task.method)
                            rpc_method = RPC_METHOD_MAPPING[method]()
                            rpc_method.agent_feature = self
                            rpc_method.request = task.request
                            rpc_method.logger = self.logger
                        except Exception as e:
                            self.logger.error("Got exception while preparing RPC task %s: %r", task.uuid, e)
                            result = RPCTaskResult(
                                uuid=task.uuid,
                                status=RPCTaskStatus.ERROR,
                                response=f"Got exception while preparing RPC task {task.uuid}: {repr(e)}"
                            )
                            await self.client.send_result(result)
                            continue

                        try:
                            async for result in rpc_method.run():
                                result.uuid = task.uuid
                                await self.client.send_result(result)

                            continue
                        except Exception as e:
                            self.logger.error('Got exception while running RPC task %s: %r', task.uuid, e)
                            result = RPCTaskResult(
                                uuid=task.uuid,
                                status=RPCTaskStatus.ERROR,
                                response=f"Got exception while running RPC task {task.uuid}: {repr(e)}"
                            )
                            await self.client.send_result(result)
                except Exception as e:
                    self.logger.error('Got exception while polling RPC task: %r', e)

                await asyncio.sleep(self.poll_interval)

        except asyncio.CancelledError:
            pass

    async def on_shutdown(self):
        if self.poll_task:
            self.poll_task.cancel()

        if self.client:
            await self.client.close()
