import asyncio
import json

import async_timeout
from logging import Logger
from typing import Callable, Optional

import aiohttp

from addr_agent import version
from addr_agent.agent_feature.mapping import RPCMethod
from addr_agent.agent_feature.models import RPCTask, RPCTaskResult, RPCTaskStatus


class AddrClient:
    logger: Logger
    session: aiohttp.ClientSession
    token: str
    result_queue: asyncio.Queue
    result_sender_task: asyncio.Task
    send_results: bool = True

    def __init__(self, logger: Logger, host: str, token: str):
        self.logger = logger
        self.session = aiohttp.ClientSession()
        self.host = host.rstrip('/')
        self.token = token

    async def start(self):
        self.result_queue = asyncio.Queue()
        self.result_sender_task = asyncio.create_task(self._result_sender())

    async def _do_request(self, method: Callable, path: str, data: Optional[str] = None, json: Optional[dict] = None):
        url = f'{self.host}/agent{path}'
        resp = await method(
            url=url,
            headers={
                'Authorization': f'Bearer {self.token}',
                'Content-Type': 'application/json',
                "User-Agent": "Addr-Agent/" + version.__version__,
            },
            data=data,
            json=json,
            timeout=10,
        )
        resp.raise_for_status()

        return resp

    async def poll_task(self) -> Optional[RPCTask]:
        resp = await self._do_request(self.session.get, "/poll/")

        if resp.status == 204:
            return None

        return RPCTask.from_json(await resp.text())

    async def send_result(self, result: RPCTaskResult):
        await self.result_queue.put(result)

    async def call_task(self, method: RPCMethod, request: Optional[str] = None) -> RPCTaskResult:
        try:
            resp = await self._do_request(
                self.session.post,
                "/call/",
                json={
                    "method": method.value,
                    "request": request,
                }
            )
            return RPCTaskResult.from_json(await resp.text())
        except Exception as e:
            return RPCTaskResult(
                status=RPCTaskStatus.ERROR,
                response=f"Failed to call RPC method {method.value}: {repr(e)}"
            )

    async def _result_sender(self):
        try:
            while self.send_results:
                results = []

                try:
                    async with async_timeout.timeout(0.5):
                        while True:
                            try:
                                result = self.result_queue.get_nowait()
                                results.append(result.to_dict())
                            except asyncio.QueueEmpty:
                                break
                            except Exception as e:
                                self.logger.exception("Got an error while processing RPC result: %r", e)
                                break
                            finally:
                                await asyncio.sleep(0.1)
                except asyncio.TimeoutError:
                    pass

                if results:
                    counter = 5
                    while counter > 0:
                        counter -= 1
                        try:
                            self.logger.info("Send results for RPC task: %s", results)
                            await self._do_request(
                                self.session.post,
                                "/result/",
                                data=json.dumps(results),
                            )
                            break
                        except Exception as e:
                            self.logger.exception("Got an error while sending result %s: %r", result, e)

                    for _ in results:
                        self.result_queue.task_done()

                    results.clear()
        except asyncio.CancelledError:
            pass

    async def close(self):
        self.send_results = False
        await self.result_queue.join()

        if self.result_sender_task:
            self.result_sender_task.cancel()

        if self.session:
            await self.session.close()
