import asyncio
import json
import shutil
import subprocess
from pathlib import Path
from typing import Generator
from urllib.parse import urlparse

import ansible_runner

from addr_agent.agent_feature.models import RPCMethodImpl, RPCTaskResult, RPCTaskStatus


class RPCAddrV1AnsibleRolesMethod(RPCMethodImpl):
    async def run(self) -> Generator[RPCTaskResult, None, None]:
        request = json.loads(self.request)
        host_public_key = request.get("host_public_key")
        addr_host = request["meta"]["addr_host"]
        ansible_host = addr_host["ssh_host"]
        ansible_port = addr_host["ssh_port"]
        ansible_user = addr_host["ssh_user"]
        ssh_private_key = request["ssh_private_key"]
        ansible_repos = request["ansible_repos"]
        roles = request["roles"]

        ansible_path = Path("ansible")
        shutil.rmtree(ansible_path, ignore_errors=True)
        ansible_path.mkdir(parents=True, exist_ok=True)
        repos_path = ansible_path / "repos"
        repos_path.mkdir(exist_ok=True)

        if host_public_key:
            yield RPCTaskResult(
                status=RPCTaskStatus.PROGRESS,
                response=f"Update host SSH public key",
            )
            authorized_keys = Path("/host/root/.ssh/authorized_keys")
            keys = set()

            if authorized_keys.exists():
                existing_keys = authorized_keys.read_text()
                keys.update(existing_keys.split("\n"))
            else:
                authorized_keys.parent.mkdir(parents=True, exist_ok=True)

            keys.add(host_public_key)
            keys = [x for x in keys if x]
            authorized_keys.write_text("\n".join(keys) + "\n")

        yield RPCTaskResult(
            status=RPCTaskStatus.PROGRESS,
            response=f"Update host SSH private key",
        )
        id_rsa = Path("/root/.ssh/id_rsa")

        if not id_rsa.exists():
            id_rsa.parent.mkdir(parents=True, exist_ok=True)

        id_rsa.write_text(f"{ssh_private_key}\n".replace('\r', ''))
        id_rsa.chmod(0o000600)

        roles_path = []
        for repo in ansible_repos:
            if repo.startswith("ssh://"):
                yield RPCTaskResult(
                    status=RPCTaskStatus.PROGRESS,
                    response=f"Get fingerprint of {repo}",
                )

                domain = urlparse(repo).netloc
                domain = domain.split("@")[1]
                parts = domain.split(":")
                domain = parts[0]
                try:
                    port = parts[1]
                except IndexError:
                    port = "22"

                _fingerprint = subprocess.Popen(["ssh-keyscan", "-p", port, "-H", domain], stdout=subprocess.PIPE)
                while _fingerprint.poll() is None:
                    await asyncio.sleep(1)

                fingerprints_path = Path("/root/.ssh/known_hosts")
                fingerprints = set()

                if fingerprints_path.exists():
                    fingerprints.update(fingerprints_path.read_text().split("\n"))
                else:
                    fingerprints_path.parent.mkdir(parents=True, exist_ok=True)

                fingerprints.add(_fingerprint.communicate()[0].decode("utf-8"))
                fingerprints = [x for x in fingerprints if x]
                fingerprints_path.write_text("\n".join(fingerprints) + "\n")

            yield RPCTaskResult(
                status=RPCTaskStatus.PROGRESS,
                response=f"Cloning repo {repo}",
            )

            try:
                repo_path = repos_path / str(hash(repo))
                roles_path.append(str(repo_path.absolute()))
                _git = subprocess.Popen(["git", "clone", repo, repo_path])
                while _git.poll() is None:
                    await asyncio.sleep(1)
            except Exception as e:
                self.logger.exception("Failed to clone repo: %r", e)
                yield RPCTaskResult(
                    status=RPCTaskStatus.ERROR,
                    response=f"Failed to clone repo {repo}: {repr(e)}",
                )
                return

            await asyncio.sleep(0.1)

        yield RPCTaskResult(
            status=RPCTaskStatus.PROGRESS,
            response=f"Will run ansible via {ansible_host}:{ansible_port}"
        )

        playbook = [
            {
                "name": "All",
                "hosts": "all",
                "roles": [x for x in roles],
            }
        ]

        inventory = {
            "ungrouped": {
                "hosts": {
                    ansible_host: {
                        "ansible_port": ansible_port,
                        "ansible_user": ansible_user,
                        **request["meta"],
                        "agent_host": self.agent_feature.config["host"],
                    }
                }
            }
        }

        project_path = ansible_path / "project"
        project_path.mkdir(exist_ok=True)
        playbook_path = project_path / "main.json"
        playbook_path.write_text(json.dumps(playbook))

        for role in roles:
            yield RPCTaskResult(
                status=RPCTaskStatus.PROGRESS,
                response=f"Run ansible for role {role}"
            )
            _, runner_object = ansible_runner.run_async(
                role=role,
                private_data_dir=ansible_path.absolute(),
                roles_path=":".join(roles_path),
                inventory=inventory,
                quiet=True,
            )
            while runner_object.rc is None:
                for event in runner_object.events:
                    yield RPCTaskResult(status=RPCTaskStatus.PROGRESS, print_datetime=False, response=event["stdout"])

        yield RPCTaskResult(status=RPCTaskStatus.COMPLETED)
